#include "..\..\common\ImageFilter.h"
#include "SettingsForm.h"
#include <limits>

using namespace FilterSpace;
using namespace System::Windows::Forms;

#define Sqr(a) ((a) * (a))

bool FilterCore(filtersettings* f, int sx, int sy, int sw, int sh, Semaphore^ updater)
{
	if      (!f)           return false;
	else if (!(f->datain)) return false;

	// BEGIN --> COMMON SUPPORT CODE (DO NOT MODIFY!)
	sx = INTERVAL_LIMITER(sx, 0, f->width - 1);
	sy = INTERVAL_LIMITER(sy, 0, f->height - 1);
	if (sw < 1) sw = f->width - sx;
	if (sh < 1) sh = f->height - sy;
	sw = INTERVAL_LIMITER(sw, 1, f->width - sx);
	sh = INTERVAL_LIMITER(sh, 1, f->height - sy);

	int sin   = f->stridein;
	int sout  = f->strideout;
	int thres = f->threshold * f->threshold;
	byte *din, *dout;
	// END --> COMMON SUPPORT CODE (DO NOT MODIFY!)

	// WRITE YOUR FILTER CODE HERE -->

	byte     bm, em, bi, ei;
	double   rm, mval, q, l, c; // quadratic, linear, constant

	byte    *surface_colors = f->surface_colors;
	byte    *object_colors  = f->object_colors;
	int      num_surfaces   = f->surface_colors_count;
	int      num_objects    = f->object_colors_count;
	int    **q_pre          = new int*[num_surfaces];
	double   sensitivity    = ((double)(f->threshold)) / 1000.0;
	double   doublemax      = (std::numeric_limits<double>::max)();

	for (int bi = 0; bi < num_surfaces; ++bi) // alapszinek
	{
		q_pre[bi] = new int[num_objects];
		for (int ei = 0; ei < num_objects; ++ei) // alapra nyomott szinek
		{
			q_pre[bi][ei] = 0;
			for (int k = 0; k < 3; ++k) // pixelcsatorna b,g,r
			{
				int surface = surface_colors[bi * 3 + k];
				int object  = object_colors[ ei * 3 + k];
				q_pre[bi][ei] += Sqr(object  - surface);
			}
		}
	}

	for (int y = 0; y < sh; ++y)
	{
		din  = f->datain  + sx * 3 + (sy + y) * sin;
		dout = f->dataout + y * sout;
		
		for (int x = 0; x < sw; ++x) 
		{
			mval = doublemax;
			
			for (bi = 0; bi < num_surfaces; ++bi) // alapszinek
			{
				byte* obj_color = object_colors;

				for (ei = 0; ei < num_objects; ++ei) // alapra nyomott szinek
				{
					byte* surf_color = surface_colors + bi * 3;
					l = c = 0.0;
					q = q_pre[bi][ei];
					for (int k = 0; k < 3; ++k) // pixelcsatorna b,g,r
					{
						int surfacech = *surf_color++;
						int objectch  = *obj_color++;
						int colorch   = din[k];
						int sc_diff   = surfacech - colorch;
						c += Sqr(sc_diff);
						l += sc_diff * (objectch - surfacech);
					}

					if (q == 0.0) 
					{
						mval = 0.0;
						bm = bi; em = ei;
						rm = 0.0;
						continue;
					}

					double r = - l / q;

					if (r < sensitivity) r = 0.0;
					if (r > 1.0) r = 1.0;

					double val = (q * r + 2 * l) * r + c;

					if (mval > val) 
					{
						mval = val;
						bm = bi; em = ei;
						rm = r;
					}
				}
			}
			*dout++ = (rm >= sensitivity) ? em + 128 : bm;
			din += 3;
		}

		// FOR PERFORMANCE MEASUREMENT
		ImageFilterBase::UpdateFilterStatus(f, sw, updater);
		if (f->nCoreKillSignal) { if (*(f->nCoreKillSignal)) break; }
	}
	// FILTER CODE END -->

	return true;
}







IntPtr ImageFilter::CreateFilterSettings(Byte* fs, bool silent)
{
	int indata = 0, outdata = 0, width = 0, height = 0;

	if (!fs) return IntPtr(0);

	bool cancelled = false;
	xml_node^ xn = gcnew xml_node( fs );

	filtersettings* f = XmlToMyStruct( xn, true );

	if (f && !silent)
	{
		FilterTemplate::SettingsForm^ form = gcnew FilterTemplate::SettingsForm(f);
		form->RefreshPreview(true);
		form->ShowDialog();
		cancelled = form->FilterCancelled;
		delete form;
	}

	if (cancelled)
	{
		delete f;
		delete xn;
		return IntPtr(0);
	}

	// CLONE "InputImage0" TO "OutputImage0"
	xml_node^ xt = nullptr;

	if (xn->XmlGetChildByName(xt, "InputImage0" )) xt->XmlGetChildValueByName("Data", indata);
	if (xn->XmlGetChildByName(xt, "OutputImage0")) xt->XmlGetChildValueByName("Data", outdata);


	// set OUTPUT_DATA settings
	if (outdata == 0 || indata == outdata)  // FIRST RUN OF CreateFilterSettings
	{
		if (f->width * f->height)
		{
			xml_node^ xt2;
			int stride = ((f->width + 3) >> 2) << 2;

			xn->XmlRemoveChildByName("OutputImage0");
			xt = gcnew xml_node( xn, "OutputImage0", "", "" );
			xn->children.Add( xt );
			//                     PinName,  width,    height,    stride, packed, interleaved, channelcount, pixelsize,                         data  )
			CreateImageSetting(xt, "Output", f->width, f->height, stride,      1,           1,            1,         8, (new Byte[stride * f->height]));
			xt->XmlGetChildByName( xt2, "ChannelsBitAllocation" );
			SetChannelsBitAllocation(xt2, 0, 8);

			// sznkioszts begyazsa az xml-be az indexek rtelmezshez
			byte   color_map[1024];
			memset(color_map, 0, 1024);
			for (int i = 0; i < f->surface_colors_count; i++) memcpy(color_map + i * 4,       f->surface_colors + i * 3, 3);
			for (int i = 0; i < f->object_colors_count;  i++) memcpy(color_map + i * 4 + 512, f->object_colors  + i * 3, 3);
			String^ colMap = stringToString(base64((unsigned char*)(color_map), 1024));
			xt2->XmlSetChildValueByName("ColorMap", colMap, false); 
			//MessageBox::Show("ColorMap created");
		}
		else
		{
			if (silent) return IntPtr(0);
		}
	}

	xn->XmlSetChildValueByName("Threshold", f->threshold);
	xn->XmlSetChildValueByName("nCore!ExecutionSize", 1); // PIXELWISE == 1 ALWAYS FOR THIS FILTER

	delete f;
	Byte* res = xn->XML_Tree_Marshal();
	delete xn;
	return (IntPtr)(void*)res;
}


filtersettings* ImageFilter::XmlToMyStruct(xml_node^ xn, bool creating)
{
	int iTmp;
	xml_node^ xntmp;

	if (! xn->XmlGetChildByName(xntmp, "InputImage0")) return 0;
	filtersettings* filset = new filtersettings;
	memset(filset, 0, sizeof(filtersettings));

	int packed, interleaved;
	if (GetOneImageSetting(xntmp, filset->width, filset->height, filset->stridein, packed, interleaved, filset->channelcountin, filset->pixelsizein, filset->datain)) 
	{
		MessageBox::Show("XmlToMyStruct - 0 - " + creating.ToString());
		delete filset;
		return 0;
	}

	//MessageBox::Show("XmlToMyStruct: In0 width=" + filset->width.ToString());

	if (! xn->XmlGetChildByName(xntmp, "InputImage1") && !creating)
	{
		MessageBox::Show("XmlToMyStruct - 1 - " + creating.ToString());
		delete filset;
		return 0;
	}

	if (GetOneImageSetting(xntmp, filset->width_colors, iTmp, iTmp, iTmp, iTmp, iTmp, iTmp, filset->datacolors))
	{
		MessageBox::Show("XmlToMyStruct - 2 - " + creating.ToString());
		delete filset;
		return 0;
	}

	//MessageBox::Show("XmlToMyStruct: In1 width=" + filset->width_colors.ToString());

	if (! xn->XmlGetChildByName(xntmp, "OutputImage0") && !creating)
	{
		MessageBox::Show("XmlToMyStruct - 3 - " + creating.ToString());
		delete filset;
		return 0;
	}

	if (GetOneImageSetting(xntmp, iTmp, iTmp, filset->strideout, iTmp, iTmp, filset->channelcountout, filset->pixelsizeout, filset->dataout))
	{
		MessageBox::Show("XmlToMyStruct - 4 - " + creating.ToString());
		delete filset;
		return 0;
	}

	// for safety reason, when outputimage is not already allocated
	filset->strideout = ((filset->width + 3) >> 2) << 2;

	int m = 0;
	if ( xn->XmlGetChildByName(xntmp, "Mask") )	xntmp->XmlGetChildValueByName("Data", m);
	filset->mask = (Byte*)(void*)m;

	xn->XmlSetChildValueByName("Threshold", 700, true);
	xn->XmlGetChildValueByName("Threshold", filset->threshold);


	filset->processing_size    = ( xn->XmlGetChildValueByName("Processing_Size", m) ) ? (int*)(void*)m : 0;
	filset->processing_ptr     = ( xn->XmlGetChildValueByName("Processing_Ptr", m) ) ? (int*)(void*)m : 0;
	filset->nCoreEnabled       = ( xn->XmlGetChildValueByName("nCore!Enabled", m) )    ? m : 0;
	filset->nCoreKillSignal    = ( xn->XmlGetChildValueByName("nCore!KillSignal", m) ) ? (int*)(void*)m : 0;
	filset->nCoreExecutionSize = ( xn->XmlGetChildValueByName("nCore!ExecutionSize", m) ) ? m : 1;

	// sznpaletta belltsa
	if (filset->datacolors)
	{
		filset->surface_colors_count = MIN(filset->width_colors, 16);
		filset->object_colors_count  = MIN(filset->width_colors, 8);
		filset->surface_colors       = filset->datacolors;
		filset->object_colors        = filset->datacolors + ((filset->width_colors * 3 + 3) / 4) * 4;

		for (int i = 1; i < filset->surface_colors_count; i++) // alapsznek megszmllsa
			if (memcmp(filset->surface_colors, filset->surface_colors + i * 3, 3) == 0)
				filset->surface_colors_count = i;

		for (int i = 1; i < filset->object_colors_count; i++)  // hlzati rteg sznek megszmllsa
			if (memcmp(filset->object_colors, filset->object_colors + i * 3, 3) == 0)
				filset->object_colors_count = i;
	}

	//MessageBox::Show(filset->surface_colors_count.ToString() + " " + filset->object_colors_count.ToString());

	if (!creating && filset->width * filset->height == 0)
	{
		delete filset;
		return 0;
	}

	return filset;
}



bool ImageFilter::CanLocalize(){ return false; }


IntPtr ImageFilter::GetSupportedFormats()
{
	xml_node ^xn, ^node = gcnew xml_node( nullptr, "", "", "" );

	node->children.Add( gcnew xml_node(node, "InputImageCount", "", "2" ) );
	node->children.Add( gcnew xml_node(node, "OutputImageCount", "", "1" ) );
	xn = gcnew xml_node( node, "InputImage0", "", "" );
	node->children.Add( xn );
	//                     PinName, width, height, stride, packed, interleaved, channelcount, pixelsize, data)
	CreateImageSetting(xn, "Input",     0,      0,      0,      1,           1,            3,        24,    0);
	xn->XmlGetChildByName( xn, "ChannelsBitAllocation" );
	SetChannelsBitAllocation(xn, 0, 8);
	SetChannelsBitAllocation(xn, 1, 8);
	SetChannelsBitAllocation(xn, 2, 8);


	xn = gcnew xml_node( node, "InputImage1", "", "" );
	node->children.Add( xn );
	//                     PinName , width, height, stride, packed, interleaved, channelcount, pixelsize, data)
	CreateImageSetting(xn, "Colors",     0,      1,      0,      1,           1,            3,        24,    0);
	xn->XmlGetChildByName( xn, "ChannelsBitAllocation" );
	SetChannelsBitAllocation(xn, 0, 8);
	SetChannelsBitAllocation(xn, 1, 8);
	SetChannelsBitAllocation(xn, 2, 8);

	xn = gcnew xml_node( node, "OutputImage0", "", "" );
	node->children.Add( xn );
	//                     PinName,  width, height, stride, packed, interleaved, channelcount, pixelsize, data)
	CreateImageSetting(xn, "Output",     0,      0,      0,      1,           1,            1,         8,    0);
	xn->XmlGetChildByName( xn, "ChannelsBitAllocation" );
	SetChannelsBitAllocation(xn, 0, 8);

	/*
	width,height = 2 --> width, height must be power of 2... = 8 --> multiply of 8
	packed, interleaved = 0, 1 (true, false)
	channelcount = 0 --> any positive number
	channelcount > 0 --> the value itself
	*/

	return (IntPtr)(void*)node->XML_Tree_Marshal();
}


// ***********************************************************************************************
// DO NOT MODIFY OR MOVE TO ELSEWHERE THESE HELPER FUNCTIONS
// THESE ARE HERE BECAUSE OF FilterCore IMPLEMENTED IN THIS FILE!!!

bool ImageFilter::FilterCore(filtersettings* f, int sx, int sy, int sw, int sh)
{
	return (::FilterCore(f, sx, sy, sw, sh, nullptr));
}

bool ImageFilter::ExecuteFilter(Byte* fs)
{ 
	bool b = false;
	if (!fs ) return false;
	xml_node^ xn = gcnew xml_node( fs );
	filtersettings* filset = XmlToMyStruct( xn, false );
	delete xn;
	if (filset)
	{
		*(filset->processing_ptr)	= 0;
		*(filset->processing_size)	= filset->width * filset->height;
		filset->nCoreExecutionSize	= 1; // SAFETY CONDITIONING
		int  cpucores				= System::Environment::ProcessorCount;
		bool parallelized			= filset->nCoreEnabled && (cpucores > 1); 
		filter_core_func f			= (::FilterCore);

		//parallelized = false;

		b = (parallelized) ? ImageFilterBase::ExecuteFilterBox(filset, f) : f(filset, 0, 0, 0, 0, nullptr); 

		delete filset;
		*(filset->processing_ptr)  = *(filset->processing_size);
		*(filset->nCoreKillSignal) = 0; // RESET SIGNAL BIT IN THE MAIN PROGRAM
	}
	return b;
}
