#pragma once

using namespace System;
using namespace System::ComponentModel;
using namespace System::Collections;
using namespace System::Windows::Forms;
using namespace System::Data;
using namespace System::Drawing;
using namespace System::Drawing::Imaging;


namespace FilterTemplate {

	public ref class SettingsForm : public System::Windows::Forms::Form
	{

	public:
		filtersettings* FilterSetting;
		bool			FilterCancelled, MouseButtonDown, init;
		Bitmap			^PreviewBitmap, ^FilteredBitmap, ^PaletteBitmap;
		Rectangle		SourceRect, DestinationRect;
		int				SWidth, SHeight, LastX, LastY, clickedColorIndex;
		bool			applyfilter;

	private: System::Windows::Forms::Label^			label1;
	private: System::Windows::Forms::PictureBox^	pictureBox1;
	private: System::Windows::Forms::Label^			label3;
	private: System::Windows::Forms::PictureBox^	Spacer;
	private: System::Windows::Forms::PictureBox^	PictureBoxColorMap;
	private: System::Windows::Forms::HScrollBar^	ScrollBarThreshold;


	public:
		SettingsForm(filtersettings* fs)
		{
			FilterSetting   = fs;
			init            = true;
			FilterCancelled = true;
			MouseButtonDown = false;
			SWidth			= fs->width;
			SHeight			= fs->height;

			InitializeComponent();

			ScrollBarThreshold->Minimum = 0;
			ScrollBarThreshold->Maximum = 1000;
			if (FilterSetting->threshold <= ScrollBarThreshold->Maximum && ScrollBarThreshold->Minimum <= FilterSetting->threshold) ScrollBarThreshold->Value = FilterSetting->threshold;

			PreviewBitmap  = gcnew Bitmap(Preview->ClientSize.Width, Preview->ClientSize.Height, PixelFormat::Format24bppRgb);
			FilteredBitmap = gcnew Bitmap(Preview->ClientSize.Width, Preview->ClientSize.Height, PixelFormat::Format8bppIndexed);
			PaletteBitmap  = gcnew Bitmap(PictureBoxColorMap->ClientSize.Width, PictureBoxColorMap->ClientSize.Height, PixelFormat::Format24bppRgb);

			ColorPalette^ cPal = FilteredBitmap->Palette;
			for (int i = 0; i < 256; i++) cPal->Entries[i] = Color::FromArgb(0, 0, 0);
			for (int i = 0; i < fs->surface_colors_count; i++) cPal->Entries[i]     = Color::FromArgb(fs->surface_colors[i * 3 + 2], fs->surface_colors[i * 3 + 1], fs->surface_colors[i * 3]);
			for (int i = 0; i < fs->object_colors_count;  i++) cPal->Entries[i+128] = Color::FromArgb(fs->object_colors[i * 3 + 2],  fs->object_colors[i * 3 + 1],  fs->object_colors[i * 3]);

			FilteredBitmap->Palette = cPal; 

			DestinationRect.X = (SWidth < Preview->ClientSize.Width) ? ((Preview->ClientSize.Width - SWidth) / 2) : 0;
			DestinationRect.Y = (SHeight < Preview->ClientSize.Height) ? ((Preview->ClientSize.Height - SHeight) / 2) : 0;
			DestinationRect.Width = (DestinationRect.X) ? SWidth : Preview->ClientSize.Width;
			DestinationRect.Height = (DestinationRect.Y) ? SHeight : Preview->ClientSize.Height;

			SourceRect.X = 0;
			SourceRect.Y = 0;
			SourceRect.Width = DestinationRect.Width;   // SHOULD BE EQUAL !!!
			SourceRect.Height = DestinationRect.Height; // SHOULD BE EQUAL !!!

			init = false;

			ScrollBar_Scroll(nullptr, nullptr);
			RefreshPalette();
		}

	protected:

		~SettingsForm()
		{
			if (components)
			{
				delete components;
			}
			delete PreviewBitmap;
			delete PaletteBitmap;
			delete FilteredBitmap;
		}


	private: System::Windows::Forms::Button^  ButtonApply;
		System::Windows::Forms::Button^  ButtonCancel;
		System::Windows::Forms::PictureBox^  Preview;
		System::ComponentModel::IContainer^  components;
	

#pragma region Windows Form Designer generated code

		void InitializeComponent(void)
		{
			this->Preview = (gcnew System::Windows::Forms::PictureBox());
			this->ButtonApply = (gcnew System::Windows::Forms::Button());
			this->ButtonCancel = (gcnew System::Windows::Forms::Button());
			this->label1 = (gcnew System::Windows::Forms::Label());
			this->ScrollBarThreshold = (gcnew System::Windows::Forms::HScrollBar());
			this->pictureBox1 = (gcnew System::Windows::Forms::PictureBox());
			this->label3 = (gcnew System::Windows::Forms::Label());
			this->Spacer = (gcnew System::Windows::Forms::PictureBox());
			this->PictureBoxColorMap = (gcnew System::Windows::Forms::PictureBox());
			(cli::safe_cast<System::ComponentModel::ISupportInitialize^  >(this->Preview))->BeginInit();
			(cli::safe_cast<System::ComponentModel::ISupportInitialize^  >(this->pictureBox1))->BeginInit();
			(cli::safe_cast<System::ComponentModel::ISupportInitialize^  >(this->Spacer))->BeginInit();
			(cli::safe_cast<System::ComponentModel::ISupportInitialize^  >(this->PictureBoxColorMap))->BeginInit();
			this->SuspendLayout();
			// 
			// Preview
			// 
			this->Preview->BorderStyle = System::Windows::Forms::BorderStyle::Fixed3D;
			this->Preview->Cursor = System::Windows::Forms::Cursors::NoMove2D;
			this->Preview->Location = System::Drawing::Point(12, 12);
			this->Preview->Name = L"Preview";
			this->Preview->Size = System::Drawing::Size(154, 154);
			this->Preview->TabIndex = 0;
			this->Preview->TabStop = false;
			this->Preview->MouseMove += gcnew System::Windows::Forms::MouseEventHandler(this, &SettingsForm::Preview_MouseMove);
			this->Preview->MouseDown += gcnew System::Windows::Forms::MouseEventHandler(this, &SettingsForm::Preview_MouseDown);
			this->Preview->Paint += gcnew System::Windows::Forms::PaintEventHandler(this, &SettingsForm::Preview_OnPaint);
			this->Preview->MouseUp += gcnew System::Windows::Forms::MouseEventHandler(this, &SettingsForm::Preview_MouseUp);
			// 
			// ButtonApply
			// 
			this->ButtonApply->Location = System::Drawing::Point(183, 143);
			this->ButtonApply->Name = L"ButtonApply";
			this->ButtonApply->Size = System::Drawing::Size(53, 23);
			this->ButtonApply->TabIndex = 2;
			this->ButtonApply->Text = L"Apply";
			this->ButtonApply->UseVisualStyleBackColor = true;
			this->ButtonApply->Click += gcnew System::EventHandler(this, &SettingsForm::ButtonApply_Click);
			// 
			// ButtonCancel
			// 
			this->ButtonCancel->Location = System::Drawing::Point(242, 143);
			this->ButtonCancel->Name = L"ButtonCancel";
			this->ButtonCancel->Size = System::Drawing::Size(51, 23);
			this->ButtonCancel->TabIndex = 3;
			this->ButtonCancel->Text = L"Cancel";
			this->ButtonCancel->UseVisualStyleBackColor = true;
			this->ButtonCancel->Click += gcnew System::EventHandler(this, &SettingsForm::ButtonCancel_Click);
			// 
			// label1
			// 
			this->label1->AutoSize = true;
			this->label1->ForeColor = System::Drawing::Color::White;
			this->label1->Location = System::Drawing::Point(180, 35);
			this->label1->Name = L"label1";
			this->label1->Size = System::Drawing::Size(114, 13);
			this->label1->TabIndex = 10;
			this->label1->Text = L"rzkkenysgi kszb (0-1000)";
			// 
			// ScrollBarThreshold
			// 
			this->ScrollBarThreshold->LargeChange = 1;
			this->ScrollBarThreshold->Location = System::Drawing::Point(183, 78);
			this->ScrollBarThreshold->Maximum = 255;
			this->ScrollBarThreshold->Name = L"ScrollBarThreshold";
			this->ScrollBarThreshold->Size = System::Drawing::Size(199, 17);
			this->ScrollBarThreshold->TabIndex = 28;
			this->ScrollBarThreshold->Value = 16;
			this->ScrollBarThreshold->Scroll += gcnew System::Windows::Forms::ScrollEventHandler(this, &SettingsForm::ScrollBar_Scroll);
			// 
			// pictureBox1
			// 
			this->pictureBox1->BackColor = System::Drawing::Color::Orange;
			this->pictureBox1->Location = System::Drawing::Point(172, 12);
			this->pictureBox1->Name = L"pictureBox1";
			this->pictureBox1->Size = System::Drawing::Size(3, 154);
			this->pictureBox1->TabIndex = 33;
			this->pictureBox1->TabStop = false;
			// 
			// label3
			// 
			this->label3->AutoSize = true;
			this->label3->ForeColor = System::Drawing::Color::White;
			this->label3->Location = System::Drawing::Point(397, 12);
			this->label3->Name = L"label3";
			this->label3->Size = System::Drawing::Size(83, 13);
			this->label3->TabIndex = 35;
			this->label3->Text = L"Colors to remap:";
			// 
			// Spacer
			// 
			this->Spacer->BackColor = System::Drawing::Color::Orange;
			this->Spacer->Location = System::Drawing::Point(391, 12);
			this->Spacer->Name = L"Spacer";
			this->Spacer->Size = System::Drawing::Size(3, 154);
			this->Spacer->TabIndex = 30;
			this->Spacer->TabStop = false;
			// 
			// PictureBoxColorMap
			// 
			this->PictureBoxColorMap->BorderStyle = System::Windows::Forms::BorderStyle::FixedSingle;
			this->PictureBoxColorMap->Location = System::Drawing::Point(400, 35);
			this->PictureBoxColorMap->Name = L"PictureBoxColorMap";
			this->PictureBoxColorMap->Size = System::Drawing::Size(330, 131);
			this->PictureBoxColorMap->TabIndex = 32;
			this->PictureBoxColorMap->TabStop = false;
			this->PictureBoxColorMap->MouseDown += gcnew System::Windows::Forms::MouseEventHandler(this, &SettingsForm::PictureBoxColorMap_MouseDown);
			this->PictureBoxColorMap->Paint += gcnew System::Windows::Forms::PaintEventHandler(this, &SettingsForm::PictureBoxColorMap_Paint);
			// 
			// SettingsForm
			// 
			this->AutoScaleDimensions = System::Drawing::SizeF(6, 13);
			this->AutoScaleMode = System::Windows::Forms::AutoScaleMode::Font;
			this->BackColor = System::Drawing::Color::SteelBlue;
			this->ClientSize = System::Drawing::Size(742, 178);
			this->Controls->Add(this->label3);
			this->Controls->Add(this->pictureBox1);
			this->Controls->Add(this->PictureBoxColorMap);
			this->Controls->Add(this->Spacer);
			this->Controls->Add(this->ScrollBarThreshold);
			this->Controls->Add(this->label1);
			this->Controls->Add(this->ButtonCancel);
			this->Controls->Add(this->ButtonApply);
			this->Controls->Add(this->Preview);
			this->FormBorderStyle = System::Windows::Forms::FormBorderStyle::FixedSingle;
			this->MaximizeBox = false;
			this->MinimizeBox = false;
			this->Name = L"SettingsForm";
			this->ShowIcon = false;
			this->ShowInTaskbar = false;
			this->StartPosition = System::Windows::Forms::FormStartPosition::CenterScreen;
			this->Text = L"Szn szegmentls sznpaletta alapjn";
			(cli::safe_cast<System::ComponentModel::ISupportInitialize^  >(this->Preview))->EndInit();
			(cli::safe_cast<System::ComponentModel::ISupportInitialize^  >(this->pictureBox1))->EndInit();
			(cli::safe_cast<System::ComponentModel::ISupportInitialize^  >(this->Spacer))->EndInit();
			(cli::safe_cast<System::ComponentModel::ISupportInitialize^  >(this->PictureBoxColorMap))->EndInit();
			this->ResumeLayout(false);
			this->PerformLayout();

		}
#pragma endregion

	public:
		void RefreshPreview(bool apply_filter)
		{
			applyfilter = apply_filter;
			if (applyfilter)
			{
				BitmapData^ bd = FilteredBitmap->LockBits( Rectangle(0, 0, FilteredBitmap->Width, FilteredBitmap->Height), ImageLockMode::ReadWrite, PixelFormat::Format8bppIndexed);
				FilterSetting->strideout = bd->Stride;
				FilterSetting->dataout = (Byte*)(void*)bd->Scan0;
				FilterSpace::ImageFilter::FilterCore(FilterSetting, SourceRect.X, SourceRect.Y, SourceRect.Width, SourceRect.Height);
				FilteredBitmap->UnlockBits(bd);
			}
			else
			{
				BitmapData^ bd = PreviewBitmap->LockBits( Rectangle(0, 0, PreviewBitmap->Width, PreviewBitmap->Height), ImageLockMode::ReadWrite, PixelFormat::Format24bppRgb);
				FilterSetting->strideout = bd->Stride;
				FilterSetting->dataout = (Byte*)(void*)bd->Scan0;
				int tmp = FilterSetting->pixelsizeout;
				FilterSetting->pixelsizeout = FilterSetting->pixelsizein;
				FilterSpace::ImageFilter::ClipImage( FilterSetting, SourceRect.X, SourceRect.Y, SourceRect.Width, SourceRect.Height);
				FilterSetting->pixelsizeout = tmp;
				PreviewBitmap->UnlockBits(bd);
			}

			this->Refresh();
		}

	private: 
		System::Void ButtonApply_Click(System::Object^  sender, System::EventArgs^  e) 
		{
			FilterCancelled = false;
			this->Close();
		}

		System::Void ButtonCancel_Click(System::Object^  sender, System::EventArgs^  e) 
		{
			this->Close();
		}

		System::Void Preview_MouseDown(System::Object^  sender, System::Windows::Forms::MouseEventArgs^  e) 
		{
			if (e->Button == System::Windows::Forms::MouseButtons::Left) 
			{
				MouseButtonDown = true;
				LastX = e->X;
				LastY = e->Y;
				RefreshPreview(false);
			}
		}

		System::Void Preview_MouseMove(System::Object^  sender, System::Windows::Forms::MouseEventArgs^  e) 
		{
			if (!MouseButtonDown) return;
			if (!DestinationRect.X)
			{
				SourceRect.X -= e->X - LastX;
				if (SourceRect.X < 0) SourceRect.X = 0;
				if (SourceRect.X >= SWidth - DestinationRect.Width) SourceRect.X = SWidth - DestinationRect.Width - 1;
			}
			if (!DestinationRect.Y)
			{
				SourceRect.Y -= e->Y - LastY;
				if (SourceRect.Y < 0) SourceRect.Y = 0;
				if (SourceRect.Y >= SHeight - DestinationRect.Height) SourceRect.Y = SHeight - DestinationRect.Height - 1;
			}
			RefreshPreview(false);
			LastX = e->X;
			LastY = e->Y;
		}

		System::Void Preview_MouseUp(System::Object^  sender, System::Windows::Forms::MouseEventArgs^  e) 
		{
			if (e->Button == System::Windows::Forms::MouseButtons::Left) MouseButtonDown = false;
			if (!MouseButtonDown) RefreshPreview(true);
		}

		System::Void Preview_OnPaint(System::Object^  sender, System::Windows::Forms::PaintEventArgs^  e) 
		{
			Graphics^ g = e->Graphics;
			Rectangle r = Rectangle(0, 0, DestinationRect.Width, DestinationRect.Height);
			if (applyfilter)
				g->DrawImage(FilteredBitmap, DestinationRect, r, GraphicsUnit::Pixel);
			else
				g->DrawImage(PreviewBitmap, DestinationRect, r, GraphicsUnit::Pixel);
		}

		void PictureBoxColorMap_Paint(System::Object^  sender, System::Windows::Forms::PaintEventArgs^  e) 
		{
			Graphics^ g = e->Graphics;
			Rectangle r = Rectangle(0, 0, PaletteBitmap->Width, PaletteBitmap->Height);
			g->DrawImage(PaletteBitmap, r, r, GraphicsUnit::Pixel);
		}


		void ScrollBar_Scroll(System::Object^  sender, System::Windows::Forms::ScrollEventArgs^  e)
		{
			if (!init)
			{
				FilterSetting->threshold = ScrollBarThreshold->Value;
			}
			label1->Text = "rzkenysgi kszb (0..1000): " + ScrollBarThreshold->Value.ToString();
			RefreshPreview(true);
		}

		void RefreshPalette()
		{
			Graphics^ gr    = Graphics::FromImage(PaletteBitmap);

			int   height      = PaletteBitmap->Height;
			int   width       = PaletteBitmap->Width;
			int   blockwidth  = width / 16;
			int   blockheight = height / 2;

			Pen^   pen   = gcnew Pen(Color::White);
			Brush^ brush = gcnew SolidBrush(Color::FromArgb(192, 192, 192));
			gr->FillRectangle(brush, Rectangle(0, 0, width, height));
			delete brush;

			for (int bx = 0; bx < FilterSetting->surface_colors_count; bx++)
			{
				byte b = FilterSetting->surface_colors[bx * 3 + 0];
				byte g = FilterSetting->surface_colors[bx * 3 + 1];
				byte r = FilterSetting->surface_colors[bx * 3 + 2];
				brush = gcnew SolidBrush(Color::FromArgb(r, g, b));
				gr->FillRectangle(brush, Rectangle(bx * blockwidth + 1, 1, blockwidth - 2, blockheight - 2));
				gr->DrawRectangle(pen, Rectangle(bx * blockwidth + 1, 1, blockwidth - 2, blockheight - 2));
				delete brush;
			}

			for (int bx = 0; bx < FilterSetting->object_colors_count; bx++)
			{
				byte b = FilterSetting->object_colors[bx * 3 + 0];
				byte g = FilterSetting->object_colors[bx * 3 + 1];
				byte r = FilterSetting->object_colors[bx * 3 + 2];
				brush = gcnew SolidBrush(Color::FromArgb(r, g, b));
				gr->FillRectangle(brush, Rectangle(bx * blockwidth + 1, blockheight + 1, blockwidth - 2, blockheight - 2));
				gr->DrawRectangle(pen, Rectangle(bx * blockwidth + 1, blockheight + 1, blockwidth - 2, blockheight - 2));
				delete brush;
			}

			PictureBoxColorMap->Refresh();			
		}

		void PictureBoxColorMap_MouseDown(System::Object^  sender, System::Windows::Forms::MouseEventArgs^  e) 
		{
			int   blockwidth  = PaletteBitmap->Width / 16;
			int   blockheight = PaletteBitmap->Height / 2;
			int   x     = e->X / blockwidth;
			int   y     = e->Y / blockheight;
			int   index = y * 16 + x;

			if (e->Button == System::Windows::Forms::MouseButtons::Left)
			{
				clickedColorIndex = index;
			}
//			RefreshPalette();
//			Refresh();
		}
};
}
